/*
    Copyright (C) 2004 Kor Nielsen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <samplerate.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>

#include "fcntl.h"

#include FUSD_H_PATH

#include <sys/soundcard.h>
#include <jack/jack.h>
#include <sys/mman.h>
#include <sys/errno.h>


#define SUPPORTED_FORMATS AFMT_S16_LE | AFMT_S16_BE | AFMT_S16_NE | AFMT_S8 | AFMT_U8;

struct circular_buffer
{
	int buffer_size;
	int alloc_size;
	int dma;
	short* buffer;
	short* wptr;
	short* rptr;
};

struct oss_instance
{
	int sample_format;
	int num_channels;
	int sample_rate;
	
	int out_total_samples_mixed;
	
	//current write info
	short* out_samples;
	short* out_sample_ptr;
	int out_samples_left_to_mix;
	int out_write_size;
	
	int frag_size;
	int frag_count;
	
	struct circular_buffer out;
	struct circular_buffer in;
	
	int in_total_samples_read;
	
	short* in_samples;
	short* in_sample_ptr;
	int in_samples_left_to_read;
	int in_read_size;
	int in_enabled;
	
	int can_block;
	
	double out_rate_ratio;
	double in_rate_ratio;
	
	SRC_STATE* out_rate_converter;
	SRC_STATE* in_rate_converter;
	
	struct fusd_file_info* in_file_info;
	struct fusd_file_info* out_file_info;
	
	struct oss_instance* next;
	
	struct fusd_file_info* polldiff;
};

struct device_info
{
	jack_client_t* client;
	jack_port_t* output_ports[2];
	jack_port_t* input_ports[2];
	int sample_rate;
	pthread_mutex_t mutex;
	int samples_sent;
	int debuglevel;
	int total_latency;
	struct oss_instance* instances;
};

// Define the DEBUG_LEVEL values used for printfs
#define DEBUG_LEVEL_CRITICAL 0
#define DEBUG_LEVEL_ERROR 1
#define DEBUG_LEVEL_WARNING 2
#define DEBUG_LEVEL_NOTICE 3
#define DEBUG_LEVEL_INFO 4
#define DEBUG_LEVEL_DEBUG1 5
#define DEBUG_LEVEL_DEBUG2 6
#define DEBUG_LEVEL_DEBUG3 7
#define DEBUG_LEVEL_DEBUG4 8

// Define the returnvalues mentionned above in help message:
#define ERR_SUCCESS 0
#define ERR_BADARG 1
#define ERR_JACKD_FAIL 2
#define ERR_FUSD_FAIL 3
#define ERR_DETACH_FAIL 4
#define ERR_JACKD_EXIT 5

// Define the maximum value for device number
#define MAX_DEVIVE_NUMBER 99

// Global variables which define debuglevel (how much data to output) or wether
// we are daemonized, so we know where to log
int debuglevel = DEBUG_LEVEL_NOTICE;
int daemonized = 0;

/**
 * Print formatted message to stdout/stderr or syslog depending on
 * importance of output and wether we are running as daemon
 * @param level
 *     Importance level, should be one of the DEBUG_LEVEL_* contants
 * @param format
 *     Message formatting rules (see printf for details)
 * @param ...
 *     Values to be formatted into message.
 */
inline void pprintf(const int level, const char *format, ...)
{
	if (level <= debuglevel)
	{
		va_list ap;
		va_start(ap, format);
		if (daemonized)
		{
			int dl;
			switch (level) {
				case DEBUG_LEVEL_CRITICAL:
					dl = LOG_CRIT;
					break;
				case DEBUG_LEVEL_ERROR:
					dl = LOG_ERR;
					break;
				case DEBUG_LEVEL_WARNING:
					dl = LOG_WARNING;
					break;
				case DEBUG_LEVEL_NOTICE:
					dl = LOG_NOTICE;
					break;
				case DEBUG_LEVEL_INFO:
					dl = LOG_INFO;
					break;
				default:
					dl = LOG_DEBUG;
			}
			vsyslog(dl, format, ap);
		}
		else
			if (level > DEBUG_LEVEL_WARNING)
				vprintf(format, ap);
			else
				vfprintf(stderr, format, ap);
		va_end(ap);
	}
}

/**
 * Print formatted message to stdout/stderr or syslog depending on
 * importance of output and wether we are running as daemon.
 * Error number (errno) will be appended as human-readable string.
 * @param level
 *     Importance level, should be one of the DEBUG_LEVEL_* contants
 * @param format
 *     Message formatting rules (see printf for details)
 * @param ...
 *     Values to be formatted into message.
 */
inline void pperror(const int level, const char *format, ...)
{
	int err = errno;
	if (level <= debuglevel)
	{
		va_list ap;
		va_start(ap, format);
		if (daemonized)
		{
			int dl;
			switch (level) {
				case DEBUG_LEVEL_CRITICAL:
					dl = LOG_CRIT;
					break;
				case DEBUG_LEVEL_ERROR:
					dl = LOG_ERR;
					break;
				case DEBUG_LEVEL_WARNING:
					dl = LOG_WARNING;
					break;
				case DEBUG_LEVEL_NOTICE:
					dl = LOG_NOTICE;
					break;
				case DEBUG_LEVEL_INFO:
					dl = LOG_INFO;
					break;
				default:
					dl = LOG_DEBUG;
			}
			vsyslog(dl, format, ap);
			vsyslog(dl, ": %s", strerror(err));
		}
		else
			if (level > DEBUG_LEVEL_WARNING)
			{
				vprintf(format, ap);
				printf(": %s\n", strerror(err));
			}
			else
			{
				vfprintf(stderr, format, ap);
				fprintf(stderr, ": %s\n", strerror(err));
			}
		va_end(ap);
	}
}

/**
 * Reallocate the buffer of OSS connection instance
 * @param instance pointer to the current oss instance
 * @param buffer_size FIXME
 */
void resize_buffer(struct circular_buffer* cbuf, int buffer_size)
{
	if(buffer_size > cbuf->buffer_size)
	{
		if(cbuf->buffer)
			munmap(cbuf->buffer, cbuf->alloc_size);
		
	cbuf->alloc_size = (buffer_size + 1) * sizeof(short);
	
	// TODO: ask for page size
	if(cbuf->alloc_size % 4096 != 0)
		cbuf->alloc_size = ((cbuf->alloc_size / 4096) + 1) * 4096;
		
		cbuf->buffer = (short*)mmap(NULL, cbuf->alloc_size, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, 0, 0);
		if(cbuf->buffer == NULL)
		{
			perror("could not allocate memory");
			exit(-1);
		}
	}
	
	// FIXME Should buffer not be shrinked when requested size is smaller than current size?
	
	cbuf->buffer_size = buffer_size;
	cbuf->wptr = cbuf->buffer;
	cbuf->rptr = cbuf->buffer;
	printf("buffer_size: %d\n", cbuf->buffer_size);
}

/**
 * Retrieves the number of samples in the buffer
 * @param instance FIXME
 * @return FIXME
 */
int cb_get_count(struct circular_buffer* cbuf)
{
	if(cbuf->dma)
		return cbuf->buffer_size;
	
	int sample_count = cbuf->wptr - cbuf->rptr;
	if(sample_count < 0)
		sample_count += cbuf->buffer_size + 1;
	return sample_count;
}

int cb_get_available(struct circular_buffer* cbuf)
{
	if(cbuf->dma)
		return cbuf->buffer_size;
	
	int sample_count = cbuf->wptr - cbuf->rptr;
	if(sample_count < 0)
		sample_count += cbuf->buffer_size + 1;
	return cbuf->buffer_size - sample_count;
}

/**
 * FIXME
 * @param instance FIXME
 * @param samples FIXME
 * @param num_samples FIXME
 * @return FIXME
 */
int cb_read(struct circular_buffer* cbuf, float* samples, int num_samples)
{
	short* s;
	float* d;
	float* end;
	
	int max_samples = cb_get_count(cbuf);
	if(num_samples > max_samples)
		num_samples = max_samples;
	
	s = cbuf->rptr;
	d = samples;
	end = samples + num_samples;
	
	while(d < end)
	{
		(*d++) = (*s++) * ( 1.0f / 32768.0f);
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
	}
	return num_samples;
}


/**
 * FIXME
 * @param instance FIXME
 * @param samples FIXME
 * @param num_samples FIXME
 * @return FIXME
 */
int cb_read_short(struct circular_buffer* cbuf, short* samples, int num_samples)
{
	short* s;
	short* d;
	short* end;
	
	int max_samples = cb_get_count(cbuf);
	if(num_samples > max_samples)
		num_samples = max_samples;
	
	s = cbuf->rptr;
	d = samples;
	end = samples + num_samples;
	
	while(d < end)
	{
		(*d++) = (*s++);
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
	}
	cbuf->rptr = s;
	return num_samples;
}


int cb_read_stereo_mix(struct circular_buffer* cbuf, float* left, float* right, int num_samples)
{
	short* s;
	float* left_end;
	
	int max_samples = cb_get_count(cbuf);
	if(num_samples * 2 > max_samples)
		num_samples = max_samples / 2;
	
	s = cbuf->rptr;
	left_end = left + num_samples;
	
	while(left < left_end)
	{
		(*left++) += (*s++) * ( 1.0f / 32768.0f);
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
		
		(*right++) += (*s++) * ( 1.0f / 32768.0f);
		
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
	}
	
	cbuf->rptr = s;
	return num_samples;
}

int cb_read_mono_mix(struct circular_buffer* cbuf, float* left, float* right, int num_samples)
{
	short* s;
	float* left_end;
	
	int max_samples = cb_get_count(cbuf);
	if(num_samples > max_samples)
		num_samples = max_samples;
	
	s = cbuf->rptr;
	left_end = left + num_samples;
	
	while(left < left_end)
	{
		(*left++) += (*s) * ( 1.0f / 32768.0f);
		(*right++) += (*s++) * ( 1.0f / 32768.0f);
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
	}
	cbuf->rptr = s;
	return num_samples;
}

/**
 * FIXME
 * @param instance FIXME
 * @param num_samples FIXME
 */
int cb_skip(struct circular_buffer* cbuf, int num_samples)
{
	short* s;
	int total_samples = cb_get_count(cbuf);
	if(num_samples > total_samples)
		num_samples = total_samples;
		
	s = cbuf->rptr;
	s += num_samples;
	if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
		s -= cbuf->buffer_size + 1 - cbuf->dma;
	
	cbuf->rptr = s;
	return num_samples;
}


int cb_write(struct circular_buffer* cbuf, float* samples, int count)
{
	int max_samples = cbuf->buffer_size - cb_get_count(cbuf);
	if(count > max_samples)
		count = max_samples;
		
	float* samples_end = samples + count;
	short* d_end = cbuf->buffer + cbuf->buffer_size + 1;
	short* d = cbuf->wptr;

	while(samples < samples_end)
	{
		(*d++) = (short) ((*samples++) * 32768.0f);
	
		if(d >= d_end)
			d = cbuf->buffer;
	}
	
	cbuf->wptr = d;
	return count;
}


int cb_write_short(struct circular_buffer* cbuf, short* s, int count)
{
	int max_samples = cbuf->buffer_size - cb_get_count(cbuf);
	if(count > max_samples)
		count = max_samples;
		
	short* s_end = s + count;
	short* d_end = cbuf->buffer + cbuf->buffer_size + 1;
	short* d = cbuf->wptr;

	while(s < s_end)
	{
		(*d++) = (*s++);
	
		if(d >= d_end)
			d = cbuf->buffer;
	}
	
	cbuf->wptr = d;
	return count;
}

int cb_write_stereo(struct circular_buffer* cbuf, float* left, float* right, int count)
{
	int max_samples = cbuf->buffer_size - cb_get_count(cbuf);
	if(count > max_samples * 2)
		count = max_samples / 2;
		
	float* left_end = left + count;
	short* d_end = cbuf->buffer + cbuf->buffer_size + 1;
	short* d = cbuf->wptr;

	while(left < left_end)
	{
		(*d++) = (short) ((*left++) * 32768.0f);
	
		if(d >= d_end)
			d = cbuf->buffer;
			
		(*d++) = (short) ((*right++) * 32768.0f);
		
		if(d >= d_end)
			d = cbuf->buffer;
	}
	
	cbuf->wptr = d;
	return count * 2;
}

int cb_write_mono(struct circular_buffer* cbuf, float* left, float* right, int count)
{
	int max_samples = cbuf->buffer_size - cb_get_count(cbuf);
	if(count > max_samples)
		count = max_samples;
		
	float* left_end = left + count;
	short* d_end = cbuf->buffer + cbuf->buffer_size + 1;
	short* d = cbuf->wptr;

	while(left < left_end)
	{
		(*d++) = (short) (((*left++) + (*right++)) * 32768.0f);
	
		if(d >= d_end)
			d = cbuf->buffer;
	}
	
	cbuf->wptr = d;
	return count;
}

void do_notify_complete(struct oss_instance* instance, struct device_info* dinfo);

/**
 * Jackd Callback -- FIXME
 * @param nframes FIXME
 * @param arg FIXME
 * @return FIXME
 */
int jack_process (jack_nframes_t nframes, void *arg)
{
	struct device_info* info = (struct device_info*) arg;
	
	float* out_left = (float*) jack_port_get_buffer (info->output_ports[0], nframes);
	float* out_right = (float*) jack_port_get_buffer (info->output_ports[1], nframes);
	
	float* in_left = (float*) jack_port_get_buffer (info->input_ports[0], nframes);
	float* in_right = (float*) jack_port_get_buffer (info->input_ports[1], nframes);

	// clear jack buffers
	{
		float* out_left_end = out_left + nframes;
		float* out_l = out_left;
		float* out_r = out_right;
		
		while(out_l < out_left_end)
		{
			*out_l++ = 0;
			*out_r++ = 0;
		}
	}
	
	pthread_mutex_lock(&info->mutex);
	
	struct oss_instance* instance = info->instances;
	while(instance != NULL)
	{
		int count = cb_get_count(&instance->out);
		
		short* s = instance->out.rptr;
		short* end = instance->out.buffer + instance->out.buffer_size + 1;
		float* out_left_end = out_left + nframes;
		float* out_l = out_left;
		float* out_r = out_right;
		
		if(instance->out_rate_converter != NULL && cb_get_count(&instance->out) > 0)
		{
			int err;
			float in[8192];
			float out[8192];
			float* s;
			SRC_DATA data;

			data.end_of_input = 0;
			data.src_ratio = instance->out_rate_ratio;
			
			data.data_in = in;
			data.data_out = out;
			data.input_frames = cb_read(&instance->out, in, 8192) / instance->num_channels;
			data.output_frames = nframes;
			
			//if(data.output_frames > nframes)
			//	data.output_frames = nframes;
			
			err = src_process(instance->out_rate_converter, &data);
			
			if(err)
			{
				printf("error: %s\n", src_strerror(err));
				break;
			}
			
			s = data.data_out;
			
			out_left_end = out_l + data.output_frames_gen;
			cb_skip(&instance->out, data.input_frames_used * instance->num_channels);

			if(instance->num_channels == 2)
			{
				while(out_l < out_left_end)
				{
					*out_l++ += *s++;
					*out_r++ += *s++;
				}
			}
			
			if(instance->num_channels == 1)
			{
				while(out_l < out_left_end)
				{
					*out_l++ += *s;
					*out_r++ += *s++;
				}
			}
			instance->out_total_samples_mixed +=  data.input_frames_used * instance->num_channels;
		}
		else
		{
			if(count > nframes)
				count = nframes;
				
			out_left_end = out_left + count;
			
			if(instance->num_channels == 2)
			{
				cb_read_stereo_mix(&instance->out, out_left, out_right, nframes);
			}
			
			if(instance->num_channels == 1)
			{
				cb_read_mono_mix(&instance->out, out_left, out_right, nframes);
			}
			instance->out_total_samples_mixed += count * instance->num_channels;
		}
		instance = instance->next;
	}	
	
	info->samples_sent += nframes * 2;
	
	instance = info->instances;
	while(instance != NULL)
	{
		if(instance->out_sample_ptr != NULL)
		{
			if(instance->out_samples_left_to_mix > 0)
			{
				int samples_written = cb_write_short(&instance->out, instance->out_sample_ptr, 
																						instance->out_samples_left_to_mix);
				
				instance->out_sample_ptr += samples_written;
				
				instance->out_samples_left_to_mix -= samples_written;
			}
			
			if(instance->out_samples_left_to_mix == 0 && cb_get_available(&instance->out) >= instance->frag_size / 2)
			{
				instance->out_sample_ptr = NULL;
				pprintf(DEBUG_LEVEL_DEBUG2, "fusd_return: %d bytes\n", instance->out_write_size);
				fusd_return(instance->out_file_info, instance->out_write_size);
				instance->out_write_size = 0;
			}
		}
		do_notify_complete(instance, info);
		instance = instance->next;
	}
	
	// Process Input
	

	instance = info->instances;
	while(instance != NULL)
	{
		if(!instance->in_enabled)
		{
			instance = instance->next;
			continue;
		}
		
		int space_available = cb_get_available(&instance->in);
		int space_needed = nframes * instance->num_channels;
		int samples_written = 0;
		
		if(instance->in_rate_converter != NULL)
		{
			int err;
			float in[8192];
			float out[8192];
			SRC_DATA data;
			
			float* d = &in[0];
			float* in_left_end = in_left + nframes;
			
			if(instance->num_channels == 2)
			{
				while(in_left < in_left_end)
				{
					*d++ = *in_left++;
					*d++ = *in_right++;
				}
			}
			
			if(instance->num_channels == 1)
			{
				while(in_left < in_left_end)
				{
					*d++ = (*in_left++) + (*in_right++);
				}
			}
			
			data.end_of_input = 0;
			data.src_ratio = instance->in_rate_ratio;
			
			data.data_in = in;
			data.data_out = out;
			data.input_frames = nframes;
			data.output_frames = 8192;
			
			//if(data.output_frames > nframes)
			//	data.output_frames = nframes;
			
			err = src_process(instance->in_rate_converter, &data);
			
			space_needed = data.output_frames_gen;
			
			if(space_needed > space_available)
				cb_skip(&instance->in, space_needed - space_available);
			
			cb_write(&instance->in, out, data.output_frames_gen);
			
		}
		else
		{
			if(space_needed > space_available)
				cb_skip(&instance->in, space_needed - space_available);
			
			if(instance->num_channels == 1)
			{
				samples_written = cb_write_mono(&instance->in, in_left, in_right, nframes);
			}
			if(instance->num_channels == 2)
			{
				samples_written = cb_write_stereo(&instance->in, in_left, in_right, nframes);
			}
		}
		
		instance->in_total_samples_read += samples_written;
		
		if(instance->in_sample_ptr != NULL)
		{
			
			int samples_read = cb_read_short(&instance->in, instance->in_sample_ptr, instance->in_samples_left_to_read);
			
			instance->in_samples_left_to_read -= samples_read;
			instance->in_sample_ptr += samples_read;
			
			if(instance->in_samples_left_to_read == 0)
			{
				instance->in_sample_ptr = NULL;
				fusd_return(instance->in_file_info, instance->in_read_size);
				instance->in_read_size = 0;
			}
		}
		
		instance = instance->next;
	}
	
	pthread_mutex_unlock(&info->mutex);
	
	return 0;
}

/**
 * FUSD callback method for "open" on device.
 * Here we setup our structures for the new connection
 * @param file FIXME
 * @return FIXME
 */
int do_open(struct fusd_file_info* file)
{
	pprintf(DEBUG_LEVEL_INFO, "Device opened...\n");
	struct oss_instance* instance = (struct oss_instance*) malloc(sizeof(struct oss_instance));
	struct device_info* info = (struct device_info*) file->device_info;
	
	instance->can_block = !(file->flags & O_NONBLOCK);
	//printf("can_block: %d\n", instance->can_block);
	
	instance->sample_format = AFMT_S16_NE;
	instance->num_channels = 2;
	instance->sample_rate = info->sample_rate;
	instance->out_rate_converter = NULL;
	instance->in_rate_converter = NULL;
	
	instance->out_samples = NULL;
	instance->out_sample_ptr = NULL;
	instance->out_samples_left_to_mix = 0;
	instance->out_total_samples_mixed = 0;
	instance->polldiff = NULL;
	
	instance->out.buffer = NULL;
	instance->out.wptr = instance->out.buffer;
	instance->out.rptr = instance->out.buffer;
	instance->out.buffer_size = 0;
	instance->out.dma = 0;
	
	instance->in_samples = NULL;
	instance->in_sample_ptr = NULL;
	instance->in_samples_left_to_read = 0;
	instance->in_enabled = 0;
	
	instance->in.buffer = NULL;
	instance->in.wptr = instance->out.buffer;
	instance->in.rptr = instance->out.buffer;
	instance->in.buffer_size = 0;
	instance->in.dma = 0;
	
	instance->frag_size = 4096;
	instance->frag_count = 4;
	
	resize_buffer(&instance->out, instance->frag_size * instance->frag_count / 2);
	resize_buffer(&instance->in, instance->frag_size * instance->frag_count / 2);
	
	file->private_data = instance;
	
	pthread_mutex_lock(&info->mutex);
	instance->next = info->instances;
	info->instances = instance;
	pthread_mutex_unlock(&info->mutex);
	
	return 0;
}

/**
 * FUSD callback method for "close" on device.
 * Here we clear our structures for the closed connection
 * @param file FIXME
 * @return FIXME
 */
int do_close(struct fusd_file_info* file)
{
	pprintf(DEBUG_LEVEL_INFO, "Device closed...\n");
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* info = (struct device_info*) file->device_info;
	
	pthread_mutex_lock(&info->mutex);
	
	instance->out.wptr = instance->out.buffer;
	instance->out.rptr = instance->out.buffer;
	
	struct oss_instance* i = info->instances;
	while(i != NULL)
	{
		if(i->next == instance)
		{
			i->next = instance->next;
			break;
		}
		i = i->next;
	}
	if(info->instances == instance)
		info->instances = instance->next;
	pthread_mutex_unlock(&info->mutex);
	
	if(instance->out_rate_converter)
		src_delete(instance->out_rate_converter);
	
	if(instance->in_rate_converter)
		src_delete(instance->in_rate_converter);
		
	if(instance->in.buffer)
		munmap(instance->in.buffer, instance->in.alloc_size);
		
	if(instance->out.buffer)
		munmap(instance->out.buffer, instance->in.alloc_size);
	
	if(instance != NULL)
		free(instance);
	
	return 0;
}

/**
 * FUSD callback method for "read" on device.
 * We should send out the data we got from jackd
 * TODO: This should get implemented!
 * @param file FIXME
 * @param buffer FIXME
 * @param size FIXME
 * @param offset FIXME
 * @return FIXME
 */
int do_read(struct fusd_file_info* file, char* buffer, size_t size, loff_t* offset)
{
	pprintf(DEBUG_LEVEL_DEBUG2, "Reading %d bytes to device..\n", size);
	
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* info = (struct device_info*) file->device_info;
	
	int sampleCount;
	if(instance->num_channels == 2)
		sampleCount = (size >> 2) << 1;
	else
		sampleCount = (size >> 1);
	
	short* s = (short*) buffer;
	
	pthread_mutex_lock(&info->mutex);
	
	instance->in_enabled = 1;
	
	int samples_read = cb_read_short(&instance->in, s, sampleCount);
	if(instance->can_block && samples_read < size)
	{
		instance->in_samples = (short*) buffer;
		instance->in_sample_ptr = s + samples_read;
		instance->in_samples_left_to_read = sampleCount - samples_read;
		instance->in_read_size = sampleCount * 2;
		instance->in_file_info = file;
		pthread_mutex_unlock(&info->mutex);
		return -FUSD_NOREPLY;
	}
	pthread_mutex_unlock(&info->mutex);
	return samples_read * 2;
}

/**
 * FUSD callback method for "write" on device.
 * We forward the data written to the device to jackd
 * @param file FIXME
 * @param buffer FIXME
 * @param size FIXME
 * @param offset FIXME
 * @return FIXME
 */
int do_write(struct fusd_file_info *file, const char *buffer,
             size_t size, loff_t *offset)
{
	pprintf(DEBUG_LEVEL_DEBUG2, "Writing %d bytes to device: %d\n", size, (file->flags & O_NONBLOCK));
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* info = (struct device_info*) file->device_info;
	
	int sampleCount;
	if(instance->num_channels == 2)
		sampleCount = (size >> 2) << 1;
	else
		sampleCount = (size >> 1);
	short* dst = (short*) buffer;
	
	pthread_mutex_lock(&info->mutex);
	
	/*
	if(cb_get_available(&instance->out) < instance->frag_size / 2)
	{
		if(instance->can_block)
		{
			instance->out_samples = (short*) buffer;
			instance->out_sample_ptr = s;
			instance->out_samples_left_to_mix = sampleCount;
			instance->out_write_size = sampleCount * 2;
			instance->out_file_info = file;
			pthread_mutex_unlock(&info->mutex);
			return -FUSD_NOREPLY;
		}
		else
		{
			pthread_mutex_unlock(&info->mutex);
			return -EAGAIN;
		}
	}*/
	
	int samples_written = cb_write_short(&instance->out, dst, sampleCount);
	
	if(!(file->flags & O_NONBLOCK) && samples_written < sampleCount)
	{
		instance->out_samples = (short*) buffer;
		instance->out_sample_ptr = dst + samples_written;
		instance->out_samples_left_to_mix = sampleCount - samples_written;
		instance->out_write_size = sampleCount * 2;
		instance->out_file_info = file;
		pthread_mutex_unlock(&info->mutex);
		return -FUSD_NOREPLY;
	}
	else if((file->flags & O_NONBLOCK) && samples_written == 0)
	{
		pthread_mutex_unlock(&info->mutex);
		return -EAGAIN;
	}
	
	pthread_mutex_unlock(&info->mutex);
	
	return samples_written * 2;
}

/**
 * FIXME
 * @param instance FIXME
 * @param dinfo FIXME
 */
void do_notify_complete(struct oss_instance* instance, struct device_info* dinfo)
{
	int curr_state, cached_state;
	
	if(instance == NULL || instance->polldiff == NULL)
	{
		return;
	}
	if(cb_get_available(&instance->out) >= instance->frag_size / 2)
		curr_state = FUSD_NOTIFY_OUTPUT;
	else
		curr_state = 0;
	
	if(cb_get_count(&instance->in) > instance->frag_size / 2)
		curr_state |= FUSD_NOTIFY_INPUT;
	
	cached_state = fusd_get_poll_diff_cached_state(instance->polldiff);
	
	if(curr_state != cached_state)
	{
		pprintf(DEBUG_LEVEL_DEBUG2, "fusd_return\n");
		fusd_return(instance->polldiff, curr_state);
		instance->polldiff = NULL;
	}
}

/**
 * FUSD callback method for "polldiff" on device.
 * FIXME
 * @param file FIXME
 * @param cached_state FIXME
 * @return FIXME
 */
ssize_t do_notify_polldiff(struct fusd_file_info* file, unsigned int cached_state)
{
	pprintf(DEBUG_LEVEL_DEBUG2, "PollDiff to device...\n");
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* info = (struct device_info*) file->device_info;
	
	if(instance == NULL)
		return -EINVAL;
	
	pthread_mutex_lock(&info->mutex);
	
	if(instance->polldiff != NULL)
	{
		fusd_destroy(instance->polldiff);
		instance->polldiff = NULL;
	}
	
	instance->polldiff = file;
	do_notify_complete(instance, info);
	
	pthread_mutex_unlock(&info->mutex);
	
	return -FUSD_NOREPLY;
}


int do_mmap(struct fusd_file_info *file, int offset, size_t length, int flags, void** addr, size_t* out_length)
{
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	*addr = instance->out.buffer;
	*out_length = instance->out.alloc_size;
	
	instance->out.dma = 1;
	
	printf("do_mmap %d, %d, %d\n", offset, length, flags);
	return 0;
}

/**
 * FUSD callback method for "ioctl" on device.
 * Here we process the OSS IOCTLs
 * @param file FIXME
 * @param request FIXME
 * @param data FIXME
 * @return FIXME
 */
int do_ioctl(struct fusd_file_info *file, int request, void *data)
{
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct device_info* dinfo = (struct device_info*) file->device_info;
	
	pthread_mutex_lock(&dinfo->mutex);
	int return_value = -1;
	switch(request)
	{
		case SNDCTL_DSP_RESET:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_RESET\n");
			instance->out.wptr = instance->out.buffer;
			instance->out.rptr = instance->out.buffer;
			//instance->samples_mixed = dinfo->samples_sent + 64 * 2;
			//instance->wptr = g_wptr + 64 * 2;
			//if(instance->wptr > g_wend)
			//	instance->wptr -= instance->buffer_size;
			return_value = 0;
			break;
		
		case SNDCTL_DSP_SETFMT:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_SETFMT\n");
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETFMTS:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_GETFMTS\n");
			*(int*)data = AFMT_S16_NE;
			return_value = 0;
			break;
		
		case SNDCTL_DSP_CHANNELS:
			{
				pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_CHANNELS  (%d)\n", *(int*)data);
				int nChannels = *(int*)data;
				if(nChannels <= 0)
					*(int*)data = 1;
				else if(nChannels > 2)
					*(int*)data = 2;
				
				*(int*)data = instance->num_channels;
				pprintf(DEBUG_LEVEL_DEBUG2,"    num_channels: %d\n", instance->num_channels);
				
				if(instance->out_rate_converter != NULL)
				{
					int err;
					src_delete(instance->out_rate_converter);
					instance->out_rate_converter = src_new(SRC_LINEAR, instance->num_channels, &err);
				}
				if(instance->in_rate_converter != NULL)
				{
					int err;
					src_delete(instance->in_rate_converter);
					instance->in_rate_converter = src_new(SRC_LINEAR, instance->num_channels, &err);
				}
			}
			return_value = 0;
			break;
			
		case SNDCTL_DSP_STEREO:
			{
				pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_STEREO  (%d)\n", *(int*)data);
				int nChannels = *(int*)data + 1;
				if(nChannels <= 0)
					nChannels = 1;
				else if(nChannels > 2)
					nChannels = 2;
				
				instance->num_channels = nChannels;
				
				if(instance->out_rate_converter != NULL)
				{
					int err;
					src_delete(instance->out_rate_converter);
					instance->out_rate_converter = src_new(SRC_LINEAR, instance->num_channels, &err);
				}
				if(instance->in_rate_converter != NULL)
				{
					int err;
					src_delete(instance->in_rate_converter);
					instance->in_rate_converter = src_new(SRC_LINEAR, instance->num_channels, &err);
				}
			}
			return_value = 0;
			break;
		
		case SNDCTL_DSP_SPEED:
			{
				pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_SPEED  (%d)\n", *(int*)data);
				int err = 0;
				int rate = *(int*)data;
				if(rate < 8000)
					rate = 48000;

				if(instance->out_rate_converter == NULL && (*(int*)data) != dinfo->sample_rate)
				{
					instance->out_rate_converter = src_new(SRC_LINEAR, instance->num_channels, &err);
					if(err)
					{
						*(int*)data = instance->sample_rate;
						return_value = 0;
					}
				}
				
				if(instance->in_rate_converter == NULL && (*(int*)data) != dinfo->sample_rate)
				{
					instance->in_rate_converter = src_new(SRC_LINEAR, instance->num_channels, &err);
					if(err)
					{
						*(int*)data = instance->sample_rate;
						return_value = 0;
					}
				}
				
				instance->out_rate_ratio = (double)dinfo->sample_rate / (double)rate;
				instance->in_rate_ratio = (double)rate / (double)dinfo->sample_rate;
				
				instance->sample_rate = rate;
				*(int*)data = rate;
			}
			return_value = 0;
			break;
			
		case SNDCTL_DSP_SYNC:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_SYNC  NOT IMPLEMENTED\n");
			return_value = 0;
			break;
		
		
		case SNDCTL_DSP_GETBLKSIZE:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_GETBLKSIZE\n");
			*((int*)data) = (instance->frag_size) * 2;
			return_value = 0;
			break;
		
		case SOUND_PCM_WRITE_FILTER:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SOUND_PCM_WRITE_FILTER  NOT IMPLEMENTED\n");
			break;
		
		case SNDCTL_DSP_POST:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_POST  NOT IMPLEMENTED\n");
			break;
			
		case SNDCTL_DSP_SUBDIVIDE:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_SUBDIVIDE  NOT IMPLEMENTED\n");
			break;
		
		case SNDCTL_DSP_SETFRAGMENT:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_SETFRAGMENT  (%d)\n", (*(int*)data) & 0xffff);
			instance->frag_size = (1 << ((*(int*)data) & 0xffff)) / 2;
			instance->frag_count = ((*(int*)data) & 0xffff0000) >> 16;
			
			if(instance->frag_count > 8)
				instance->frag_count = 8;
			
			(*(int*)data) = ((*(int*)data) & 0xffff ) | instance->frag_count << 16;
			
			pprintf(DEBUG_LEVEL_DEBUG1, "    FragSize: %d x %d\n", instance->frag_size, instance->frag_count);
			resize_buffer(&instance->out, instance->frag_size * instance->frag_count / 2);
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETOSPACE:
		{
			pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SNDCTL_DSP_GETOSPACE\n");
			struct audio_buf_info* info = (struct audio_buf_info*) data;
			info->fragments = cb_get_available(&instance->out) / (instance->out.buffer_size / instance->frag_count);
			if(info->fragments > instance->frag_count)
				info->fragments = instance->frag_count;
			info->fragstotal = instance->frag_count;
			info->fragsize = (instance->out.buffer_size / instance->frag_count) * 2;
			info->bytes = (instance->out.buffer_size - cb_get_count(&instance->out)) * 2;
			pprintf(DEBUG_LEVEL_DEBUG2, "    Fragstotal: %d\n    Fragsize: %d\n    Bytes: %d\n", info->fragstotal, info->fragsize, info->bytes);
			return_value = 0;
			break;
		}

		case SNDCTL_DSP_GETISPACE:
			pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SNDCTL_DSP_GETISPACE\n");
			struct audio_buf_info* info = (struct audio_buf_info*) data;
			
			info->fragments = cb_get_count(&instance->in) / (instance->out.buffer_size / instance->frag_count);
			if(info->fragments > instance->frag_count)
				info->fragments = instance->frag_count;
			info->fragstotal = instance->frag_count;
			info->fragsize = (instance->in.buffer_size / instance->frag_count) * 2;
			info->bytes = cb_get_count(&instance->in);
			return_value = 0;
			break;
		
		case SNDCTL_DSP_NONBLOCK:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_NONBLOCK  NOT IMPLEMENTED\n");
			break;
		
		case SNDCTL_DSP_GETCAPS:
			*(int*)data = DSP_CAP_TRIGGER | DSP_CAP_MMAP;
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_GETCAPS (%d)\n", *(int*)data);
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETTRIGGER:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_GETTRIGGER  NOT IMPLEMENTED\n");
			break;
		
		case SNDCTL_DSP_SETTRIGGER:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_SETTRIGGER\n");
			return_value = 0;
			break;
			
		case SNDCTL_DSP_GETIPTR:
			pprintf(DEBUG_LEVEL_DEBUG2,  "IOCTL: SNDCTL_DSP_GETIPTR\n");
			
			((count_info*)data)->bytes = instance->in_total_samples_read * 2;
			((count_info*)data)->blocks = 0;
			((count_info*)data)->ptr = 0;
			
			return_value = 0;
			break;
			
		case SNDCTL_DSP_GETOPTR:
			pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SNDCTL_DSP_GETOPTR\n");
			((count_info*)data)->bytes = instance->out_total_samples_mixed * 2 + jack_port_get_total_latency(dinfo->client, dinfo->output_ports[0]) * instance->num_channels * 2;
			
			((count_info*)data)->blocks = instance->out_total_samples_mixed * 2 / instance->frag_size;
			((count_info*)data)->ptr = (instance->out.rptr - instance->out.buffer) * 2;
			
			return_value = 0;
			break;

		case SNDCTL_DSP_MAPINBUF:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_MAPINBUF  NOT IMPLEMENTED\n");

			break;
			
		case SNDCTL_DSP_MAPOUTBUF:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_MAPOUTBUF  NOT IMPLEMENTED\n");
			break;
			
		case SNDCTL_DSP_SETSYNCRO:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_SETSYNCRO  NOT IMPLEMENTED\n");
			break;
			
		case SNDCTL_DSP_GETODELAY:
			*((int*)data) = cb_get_count(&instance->out) * 2 + jack_port_get_total_latency(dinfo->client, dinfo->output_ports[0]) * instance->num_channels * 2;
			return_value = 0;
			pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SNDCTL_DSP_GETODELAY bytes=%d\n", *((int*)data));
			break;
			
		case SNDCTL_DSP_GETCHANNELMASK:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_GETCHANNELMASK  NOT IMPLEMENTED\n");
			break;
			
		case SNDCTL_DSP_BIND_CHANNEL:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_BIND_CHANNEL  NOT IMPLEMENTED\n");
			break;
		
		default:
			pprintf(DEBUG_LEVEL_WARNING, "IOCTL: Unknown IOCTL %d,  NOT IMPLEMENTED\n", request);
			break;
	}
	
	pthread_mutex_unlock(&dinfo->mutex);
	
	return return_value;
}

struct fusd_file_operations oss_file_ops = 
{
	open: do_open,
	read: do_read,
	write: do_write,
	close: do_close,
	ioctl: do_ioctl,
	mmap: do_mmap,
	poll_diff: do_notify_polldiff
};

/**
 * Jackd Callback -- FIXME
 * @param nframes FIXME
 * @param arg FIXME
 * @return FIXME
 */
int jack_srate (jack_nframes_t nframes, void *arg)
{
	struct device_info* info = (struct device_info*) arg;
	info->sample_rate = nframes;
	return 0;
}

/**
 * Jackd Callback -- FIXME
 * @param nframes FIXME
 * @param arg FIXME
 * @return FIXME
 */
int  jack_buffer_size (jack_nframes_t nframes, void *arg)
{
	struct device_info* info = (struct device_info*) arg;
	printf("unexpected set of jack buffer size = %d\n", nframes);
	return -1;
}

/**
 * Jackd Callback -- FIXME
 * @param desc FIXME
 */
void jack_error (const char *desc)
{
	pprintf(DEBUG_LEVEL_ERROR, "JACK error: %s\n", desc);
}

/**
 * Jackd Callback -- FIXME
 * @param arg FIXME
 */
void jack_shutdown (void *arg)
{
	exit(ERR_JACKD_EXIT);
}

/**
 * FIXME
 */
void setscheduler(void)
{
	struct sched_param sched_param;

	if (sched_getparam(0, &sched_param) < 0) {
		pprintf(DEBUG_LEVEL_WARNING, "Scheduler getparam failed...\n");
		return;
	}
	sched_param.sched_priority = 10;
	if (!sched_setscheduler(0, SCHED_FIFO, &sched_param)) {
		pprintf(DEBUG_LEVEL_INFO, "Scheduler set to Round Robin with priority %i...\n", sched_param.sched_priority);
		return;
	}
	pprintf(DEBUG_LEVEL_WARNING, "!!!Scheduler set to Round Robin with priority %i FAILED!!!\n", sched_param.sched_priority);
}

/**
 * Displays help message. Calls exit() with value given in "result"
 * once information is printed.
 * @param version_only
 *    If this parameter is non-zero, only version will be displayed
 * @param result
 *    Return value to use when exiting at end of this function
 */
void print_help(int version_only, int result)
{
	printf(PACKAGE" version "VERSION"\n"
	       " 2004 Kor Nielsen\n");
	if (!version_only) {
		// Print usage information
		printf("\nUsage:\n"
		       "   "PACKAGE" -h|v\n"
		       "   "PACKAGE" [-d] [-q] [-V] [-n <int>]\n"
		       "\n"
		       "Options:\n"
		       "     -h   --help\n"
		       "        Display this help message.\n"
		       "     -v   --version\n"
		       "        Display program version.\n"
		       "     -d   --detach\n"
		       "        Detach from console when we have our connection\n"
		       "        to Jack server and FUSD. Once detached logging\n"
		       "        will happen to syslog using daemon facility.\n"
		       "     -n <int>   --devnum <int>\n"
		       "        Use device with number <int> instead of without\n"
		       "        any numbering. Allowed is -1 .. 99.\n"
		       "        E.g. If <int> = 5, then we will setup /dev/dsp5\n"
		       "        instead of /dev/dsp\n"
		       "     -q   --quiet\n"
		       "        Be more quiet\n"
		       "     -V   --verbose\n"
		       "        Be more verbose\n"
		       "\n"
		       "Exit values:\n"
		       "      0  Success in execution.\n"
		       "      1  Invalid commandline arguments.\n"
		       "      2  Failed to connect to jackd.\n"
		       "      3  Failed to setup our device(s) under /dev/.\n"
		       "      4  Failed to detach from console\n"
		       "      5  Termination because we lost connection to jackd\n");
	}
	exit(result);
}

/**
 * Parses commandline arguments
 * @param argc Count of arguments
 * @param argv Array of string containing the arguments
 * @param detach OUT: will be set to 1 if we should detach from terminal
 * @param dspnum OUT: will be set to the device number to use for /dev/dsp%d
 */
void parse_arguments(int argc, char*argv[], int*detach, int*dspnum)
{
	int i = 0;
	*detach = 0; // Default, do not detach from console
	*dspnum = -1; // Default value, no device number!
	while (argv[++i])
	{
		if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0)
		{
			print_help(0, ERR_SUCCESS);
		}
		else if (strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "--version") == 0)
		{
			print_help(1, ERR_SUCCESS);
		}
		else if (strcmp(argv[i], "-d") == 0 || strcmp(argv[i], "--detach") == 0)
		{
			*detach = 1;
		}
		else if (strcmp(argv[i], "-q") == 0 || strcmp(argv[i], "--quiet") == 0)
		{
			if (debuglevel > 0) debuglevel--;
		}
		else if (strcmp(argv[i], "-V") == 0 || strcmp(argv[i], "--verbose") == 0)
		{
			if (debuglevel < 99) debuglevel++;
		}
		else if (strcmp(argv[i], "-n") == 0 || strcmp(argv[i], "--devnum") == 0)
		{
			if (argv[++i] && strspn(argv[i], "0123456789") == strlen(argv[i]))
			{
				sscanf(argv[i], "%d", dspnum);
				if (*dspnum < -1 || *dspnum > MAX_DEVIVE_NUMBER)
				{
					printf("Device number is out of range.\n");
					print_help(0, ERR_BADARG);
				}
			}
			else
			{
				// We got no device number or garbage
				printf("Invalid device number, expecting integer in range is -1 .. %d\n", MAX_DEVIVE_NUMBER);
				print_help(0, ERR_BADARG);
			}
		}
		else
		{
			printf("Unknown option '%s'\n", argv[i]);
			print_help(0, ERR_BADARG);
		}
	} // End while
}

/**
 * Program entry-point
 * @param argc Count of arguments
 * @param argv Array of string containing the arguments
 * @return Program return value
 */
int main(int argc, char *argv[])
{
	int dspnum, detach;

	// Allow users to provide some options
	parse_arguments(argc, argv, &detach, &dspnum);

	mlockall(MCL_CURRENT | MCL_FUTURE);
	setscheduler();
	
	jack_client_t* client;
	struct device_info* info;
	const char** ports;
	
	jack_set_error_function(jack_error);
	
	client = jack_client_new("oss-dsp");
	
	if(client == NULL)
	{
		pprintf(DEBUG_LEVEL_CRITICAL, "Error registering jack client. Make sure that jackd is running.\n");
		exit(ERR_JACKD_FAIL);
	}
	
	info = (struct device_info*) malloc(sizeof(struct device_info));
	
	info->client = client;
	
	pthread_mutex_init(&info->mutex, NULL);
	info->instances = NULL;
	info->samples_sent = 0;

	info->output_ports[0] = jack_port_register(client, "out_left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	info->output_ports[1] = jack_port_register(client, "out_right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	
	info->input_ports[0] = jack_port_register(client, "in_left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	info->input_ports[1] = jack_port_register(client, "in_right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	
	
	jack_set_process_callback(client, jack_process, info);
	jack_set_sample_rate_callback(client, jack_srate, info);
	jack_set_buffer_size_callback(client, jack_buffer_size, info);
	
	jack_on_shutdown(client, jack_shutdown, info);
	
	info->total_latency = jack_port_get_total_latency(client, info->output_ports[0]);

	jack_activate(client);

	ports = jack_get_ports (client, NULL, NULL, JackPortIsPhysical|JackPortIsInput);

	jack_connect (client, jack_port_name(info->output_ports[0]), ports[0]);
	jack_connect (client, jack_port_name(info->output_ports[1]), ports[1]);
	
	free(ports);
	
	ports = jack_get_ports (client, NULL, NULL, JackPortIsPhysical|JackPortIsOutput);

	jack_connect (client, ports[0], jack_port_name(info->input_ports[0]));
	jack_connect (client, ports[1], jack_port_name(info->input_ports[1]));

	free(ports);
	
	int dsp_fd = 0;
	char device[20];
	if (dspnum >= 0)
	{
		snprintf(device, sizeof(device), "/dev/sound/dsp%d", dspnum);
	}
	else
	{
		strcpy(device, "/dev/sound/dsp");
	}
	pprintf(DEBUG_LEVEL_INFO, "Creating audio device '%s'\n", device);
#ifdef FUSD_KOR_HACKED_VERSION
	if(dsp_fd = fusd_register(device, "sound", device+11, 0666, info, &oss_file_ops) < 0)
#else
	if(dsp_fd = fusd_register(device, 0666, info, &oss_file_ops) < 0)
#endif
	{
		pperror(DEBUG_LEVEL_CRITICAL, "Could not create sound device\n");
		exit(ERR_FUSD_FAIL);
	}

	if (detach)
	{
		if (daemon(0, 0) == -1)
		{
			// We failed to daemonize...
			pperror(DEBUG_LEVEL_CRITICAL, "Could not detach from terminal\n");
			exit(ERR_DETACH_FAIL);
		}
		else
		{
			openlog(PACKAGE, 0, LOG_DAEMON);
			daemonized = 1;
		}
	}

	pprintf(DEBUG_LEVEL_INFO, "Running...\n");
	fusd_run();
	
	fusd_unregister(dsp_fd);
	pthread_mutex_destroy(&info->mutex);
	return 0;
}
